/*
 * bobpp/bob_cost.h
 * 
 * This source include the abstract data type the costed notion in Bob++
 */

#include <string>


#ifndef BOBCOST_HEADER
#define BOBCOST_HEADER
namespace Bob {


/** @defgroup CostFunctGRP  Costed functionnality.
 *  @brief The module describes all the classes used when a node of the search is costed.
 *  @ingroup SchedAlgoGRP
    In many case of exact algorithm, a node of the space search is costed.
    The cost means here value obtained with the part of variables that
    have been fixed. This is not an evalution. This is not an evaluation
    of the cost of the solutions that could be reach.

    Then we defined here all classes that should be used in this case.
 */

/** The class that represents the cost of a node
 * The type of the cost could be int, long float, double.
 * But also some user defined types.
 * @ingroup CostFunctGRP
 */
template <class T=int, class Compare = Min<T> >
class  Cost {
public :
/// Type declaration of a Cost with a specific Type t and comparison class Compare.
typedef Cost<T,Compare> TheCost;
/// Type declaration of a Cost with a specific Type t and comparison class Compare.
typedef Compare Comparator;
/// Type of the cost.
typedef T CostType;
/// The value of the cost.
T cost;
/// Constructor  the default cost value if 0.
   Cost() { cost=(T)0;}
   /// Constructor with a initialization of the cost.
   Cost(T p) { cost=p; }
   /// destructor
   virtual ~Cost() {}
   /// Method to change the cost value.
   void setCost(T t) { cost=t; }
   /// Method to returns the value of the cost.
   T getCost() const { return cost; }
   /// Operator to compare two costs.
   bool operator==(const TheCost&t1) { return Compare::isEq(cost,t1.cost);}
   /// Operator to compare two costs.
   bool operator>(const TheCost&t1) { return Compare::isBetter(cost,t1.cost);}
   /// Operator to compare two costs.
   bool operator<=(const TheCost&t1) { return !Compare::isBetter(cost,t1.cost);}
   /// copy method.
   virtual void copy(const TheCost &te) { cost=te.cost; }
   /// Affectation operator.
   const TheCost &operator=(const TheCost&t1) { copy(t1); return *this; }
   /** Pack method to serialize the SchedInstance
     * @param bs the Serialize object
     */
   virtual void Pack(Serialize &bs) const  {
         DBGAFF_ENV("Cost::Pack","Packing Instance");
         bs.Pack(&cost,1);
   }
   /** Unpack method to deserialize the SchedInstance
     * @param bs the DeSerialize object
     */
   virtual void UnPack(DeSerialize &bs)  {
         DBGAFF_ENV("Cost::UnPack","Unpacking Instance");
         bs.UnPack(&cost,1);
   }
};


};
#endif

