
/* 
 * bobpp/bob_dspq.h
 * 
 * This source include the abstract and real data types to implement 
 * the data structures that store the pending nodes. 
 */


#ifndef DSPQ_HEADER
#define DSPQ_HEADER

/*----------- Set of Nodes ----------*/

namespace  Bob {
/** @defgroup PQGRP Generic classes for Priority Queue
 *  this module includes all generic definitions for priority Queue
 *  @ingroup GDS_GRP
 * @{
*/

/** @defgroup GenePQ Generic classes for Priority Queue
 *  this module includes all generic definitions for priority Queue
 * @{
*/

/** The Priority Queue Statisitic class
 */
class PQStat : public stat {
	public:
		/// Constructor
		PQStat(bool g=false):stat('Q') { 
         add_counter('i',"Inserted Nodes",g);
         add_counter('d',"Deleted Nodes",g);
         //add_counter("l");
      }
		/// Destructor
		virtual ~PQStat() {}
		/// counts adding one node to the PQ
		virtual void Ins(long l=0) { 
			get_counter('i')++; 
         //get_counter("l")+=l;
		}
		/// counts removing one node from the PQ
		virtual void Del(long l=0) { 
			get_counter('d')++; 
         //get_counter("l")-=l;
		}
		/// counts removing nb nodes from the PQ
		virtual void DelG(int nb,long l=0) { 
		};
		/// get the number of nodes
		int GetNbN() { return get_counter('i').get()-get_counter('d').get(); }
};

/** The Priority Queue interface
 */
template<class Node, class PriComp>
class PQInterface {
public :
   /// Constructor (mainly do nothing)
   PQInterface() {}
   /// Destructor (mainly do nothing)
   virtual ~PQInterface() {}
   /// Remove a reference on the Priority Queue:w
   virtual void remRef()=0; 
   /// Add a reference on the Priority Queue:w
   virtual void addRef()=0; 
   /// Resets the Data structure
   virtual void Reset()=0; 
   /// Inserts a Node n with its associated priority p
   virtual void Ins(Node *n)=0;
   /// Deletes the best prioritized Node 
   virtual Node *Del()=0; 
   /// Returns the Best node
   virtual int DelG() =0;
   /// Prints the contents of the PQ.
   virtual ostream &Prt(ostream &os=std::cout) const =0;
   /// Get the Statistics of the priotiy queue
   virtual PQStat *getStat()=0;
};
//@}


/** @defgroup ALPQGRP Abstract Local Priority Queue
 *  This module includes definition of generic priority Queue and some implementation
 *  Skew Heap, LinkedList, etc...
 *  @ingroup PQGRP
 */
//@{

/**
   The internal node used to store the Pri with its associated Node
  */
template<class Node>
class PQNode {
protected :
  /// the node i.e. the Node (not the node of the priority queue, but the node of the 
  //  search space.
  Node *b;
public:
     /// constructor (with a priority and a Node).
     PQNode(Node *bp): b(bp) { }
     /// constructor of an empty PQNode.
     PQNode():b(0) { }
     /// Destructor
     ~PQNode() { if(b!=0 ) delete b;}
     ///  gets the associated Node.
     Node *&GetNode() { return b; }
};


/**
  The priority Queue data structure stores Nodes according to Pri
*/
template<class Node,class PriComp>
class PQ :public PQInterface<Node,PriComp> { 
PQStat pqs;
protected:
/** this data counts the number of algorithms that use this
 * priority queue at the same time.
 */
int ref;
public:
   /// constructor
   PQ(bool g=false) :PQInterface<Node,PriComp>(),pqs(g),ref(0) { }
   /// destructor
   virtual ~PQ() { BOBASSERT(ref==0,ReferenceException,"PQ","The PQ is deleted and ref!=0"); }
   /// init the reference for the internal statistics
   void Init(Id *s) { pqs.init(s); }
   /// Remove a reference on the Priority Queue:w
   virtual void remRef() {ref--;}
   /// Add a reference on the Priority Queue:w
   virtual void addRef() {ref++;}
   /// Statistics for insert.
   void StIns(Node *n=0) { pqs.Ins(); }
   /// Statistics for delete.
   void StDel(Node *n=0) { pqs.Del(); }
   /// Statistics for deleteGreater.
   void StDelG(int nb=1) { pqs.DelG(0); }
   /// Statistics for Insert by Load Balancer.
   //void StIns4Load(SetNode<T,Compare> *n=0) { bs->Set(3); }
   /// Statistics for Delete by Load Balancer.
   //void StDel4Load(SetNode<T,Compare> *n=0) { bs->Set(4); }
   /// Inserts a Node n with its associated priority p
   virtual void Ins(Node *n) {};
   /// Reset the Data structure
   virtual void Reset() { pqs.reset();}
   /// Deletes the best prioritized Node 
   virtual Node *Del() { return 0;};
   /// Deletes all the Node that have a priority less than p
   virtual int DelG() { return 0;};

   /// returns the Load of the PQ
   virtual int Load() { return 0;}

   /// inserts all the Node stored in the SetNode sn
   //virtual void Ins4Load(SetNode<T,Compare> *sn) {};
   /// deletes Node representing the load l and gets them in sn
   //virtual void Del4Load(Load *l,SetNode *sn) {};
   /// Used as is this data structure is not parallel
   bool IsParallel() { return false; }
   /// Get the Statistics of the Solution
   PQStat *getStat() { return &pqs; }
};

//@}




/** @defgroup SkewHeapGRP Skew-Heap implemantation
    @name a PQ implementations of Priority Queue : the Skew Heap.
*/

//@{

/**
  * Priority Queue implementation :  Skew heap
  */
template<class Node>
class NodeSkew : public PQNode<Node> {
  NodeSkew<Node> *left,*right;
public:
   /// constructor
   NodeSkew(Node *bp) : PQNode<Node>(bp) , left(0), right(0)
     { }
   /// constructor
   NodeSkew() : PQNode<Node>(),left(0),right(0) {}
   /// right child
   NodeSkew<Node> *&Right() { return right;}
   /// left child
   NodeSkew<Node> *&Left() {  return left;}
};

/**
  * Priority Queue implementation the Skew Heap
  */
template<class Node,class PriComp>
class PQSkew : public PQ<Node,PriComp> {
public: 
/// the type for the internal node of the Skew-Heap
typedef NodeSkew<Node> TheNode;
private:
  /// the root of the tree
TheNode *r;

   /// the meld method
   void Meld(TheNode *q1,TheNode *q2){
  	TheNode *p;
  
  	if (q1==0) { r=q2;  return; }
  	if (q2==0) { r=q1;  return; }

  	Swap(&q1, &q2);
  	r=q1;
  	p=q1;
  	q1 = p->Right();
  	p->Right() = p->Left();

  	while (q1!=0) {
      		Swap(&q1, &q2);
      		p->Left()=q1;
      		p=q1;
      		q1=p->Right();
      		p->Right()=p->Left();
    	}
	
  	p->Left()=q2;
   }
   /// the swap method
   void Swap(TheNode **q1, TheNode **q2) {
    TheNode *dum;
       if ( *q2== 0) return;
       //std::cout << "Q1:"<<*(*q1)->Pri()<<" Q2:"<< *(*q2)->Pri()<<std::endl;
       if ( PriComp::isBetter(*((*q2)->GetNode()),*((*q1)->GetNode()))  ) {
          dum = *q1;
          *q1 = *q2;
          *q2 = dum;
        }
       //std::cout << "Q1:"<<*(*q1)->Pri()<<" Q2:"<< *(*q2)->Pri()<<std::endl;
   };

public:

   /// constructor
   PQSkew(bool g=false) : PQ<Node,PriComp>(g) { r=0; }
   /// destructor
   virtual ~PQSkew() {  
      Node *f;
      while ( (f=Del())!=0 ) {
	   if ( f->isDel() ) delete f;
      }
   }
   /// Reset the priotiy queue
   void Reset() {
     Node *f;
     while ( (f=Del())!=0 ) {
	   if ( f->isDel() ) delete f;
     }
     PQ<Node,PriComp>::Reset();
   }
   /// Insert a Node n with the priority p
   void Ins(Node *n) {
	   TheNode *p = new TheNode(n);
	   n->addRef();
   	PQ<Node,PriComp>::StIns();
   	Meld(p,r);
   }
   /// Delete the Node with the highest priority
   Node *Del(){
	TheNode *p;
	Node *b;

   	p = r;
   	if ( p==0 ) return 0 ;
   	Meld(p->Right(),p->Left());
   	b = p->GetNode();
   	p->GetNode()=0;
   	PQ<Node,PriComp>::StDel( );
   	delete p;
	   b->remRef();
   	return b;
   }
   /// Delete the Nodes with a lower priority than p
   int DelG() {return 0;}
   /// Displays statistics and Contents
   ostream &Prt(ostream &o=std::cout) const { o << "Pas d'affichage\n";return o; }
};

//@}

/** @defgroup DBLKPQGRP doubled linked list implementaiton of a Priotiy Queue.
 *  @name  Another implementation of Priority queue, represented by a double linked list.
*/

//@{
/**
  * Priority Queue implementation : doubled linked list 
  */
template<class Node>
class NodePQSL : public PQNode<Node> {
   NodePQSL *ne,*pr;
public:
   /// constructor
   NodePQSL() : PQNode<Node>() { ne=0;pr=0; }
   /// constructor
   NodePQSL(Node *bp) : PQNode<Node>(bp) { ne=0;pr=0; }
   /// Next of the Label;
   NodePQSL *&N() {return ne; }
   /// Previos of the Label;
   NodePQSL *&P() {return pr; }

};

/**
  * Priority Queue implementation : doubled linked list 
  */
template<class Node,class PriComp>
class PQSL : public PQ<Node,PriComp> {
public:
/// the type for the internal node of the double linkek list
typedef NodePQSL<Node> TheNode;
protected:
/// the pointer on the previous node
TheNode *r;
/// the pointer on the next node
TheNode *l;
/// the max number of nodes stored
int Max;
/// the current number nodes
int Nb;

public:
   /** Constructor 
     */
   PQSL(bool g=false) : PQ<Node,PriComp>(g) { Max=-1;r=0;l=0; Nb=0; }
   /** Constructor 
     * @param PMax indicates the mximum number of solutions 
     *        that could be stored. 
     */
   PQSL(int PMax) : PQ<Node,PriComp>() { Max=PMax;r=0;l=0; Nb=0; }
   /// destructor
   virtual ~PQSL(){  
      Node *f;
      while ( (f=Del())!=0 ) {
         delete f;
      }
   }
   /// Reset the priotiy queue
   void Reset(){
      Node *f;
      while ( (f=Del())!=0 ) {
         delete f;
      }
      PQ<Node,PriComp>::Reset();
   }
   /// Insert a Node n with the priority p
   void Ins(Node *b){
      TheNode *c;
      TheNode *n = new TheNode(b);
      for (c=r;c!=0 && PriComp::isBetter(*(c->GetNode()),*(n->GetNode())) ;c = c->N());
      if ( c== 0 ) {
         if ( l!=0 ) {
            l->N() = n; n->P() = l;
         }
         l = n;
         if ( r==0 ) r = n;
      } else {
         n->N() = c;
         n->P()= c->P();
         c->P() = n;
         if ( n->P()==0 ) {
            r = n;
         } else {
            n->P()->N() = n;
         }
      }
      PQ<Node,PriComp>::StIns();
   }
   /// Delete the Node with the highest priority
   Node *Del(){
      TheNode *f;
      Node *b;
      f = r;
      if ( f==0 ) return 0;
      r = f->N();
      if ( r==0 ) l=0;
      else r->P() = 0;
      b = f->GetNode();
      f->GetNode()=0;
      PQ<Node,PriComp>::StDel(b);
      delete f;
      return b;
   }
   // Delete the Nodes with a lower priority than p
   int DelG() {return 0;}
   /// Displays statistics and Contents
   ostream &Prt(ostream &o=std::cout) const {  
      return o << "No Display for the PQSL at this time\n";
   }
};

//@}

//@}


};
// end of namespace.

#endif
