/*
 * Bobpp/src/bob_schedalgo.h
 * 
 * This source include the abstract data type for the Algo level where
 * the goal is to find axect solution of a problem.
 */

#include <string>


#ifndef BOBSCHEDALGO_HEADER
#define BOBSCHEDALGO_HEADER
namespace Bob {

/*---------------------------------------------------------------------*/
/** @defgroup SchedAlgoGRP Scheduled Algorithm
 *  @ingroup UserAlgoGRP
 *  @brief Algorithm interface when the search is scheduled
 * @{
 *  When the goal of the algorithm is to perform a search in a space,
 *  the space could be explorated using different strategies. The first step is then 
 *  to allow to algorithm to schedule the search i.e. to schedule the nodes. 
 *  In this case the algorithm uses
 *  -# a notion of Priority associated which each node (see @ref PriGRP)
 *  -# a priority Queue (see @ref PQGRP).
 *
 *  
 */

/*---------------------------------------------------------------------*/

/** the class that defines a Node where the search is scheduled
 * This node inherits from the Node class.
 * The functionality that is added is a boolean to know if the node is a solution to the 
 * problem or a subproblem.
 */
class SchedNode :public base_Node {
   protected:
      /// boolean solution or not.
      int b;  
   public :
      /// Constructor
      SchedNode() : base_Node(),b(0) { 
         DBGAFF_MEM("SchedNode::SchedNode","Constructor (no parameter)");
      }
      /** Constructor 
        * @param sn a SchedNode to copy
        */
      SchedNode(SchedNode const *sn):base_Node((base_Node *)sn),b(sn->b) { 
         DBGAFF_MEM("SchedNode::SchedNode","Constructor (para Node *)");
      }
      /** Constructor 
        * @param sn a SchedNode to copy
        */
      SchedNode(SchedNode const &sn):base_Node((base_Node &)sn),b(sn.b) { 
         DBGAFF_MEM("SchedNode::SchedNode","Constructor (para Node &)");
      }
      /// Destructor
      virtual ~SchedNode() {
         DBGAFF_MEM("SchedNode::~SchedNode","Destructor");
      }
      /// copy method used by the operator=
      virtual void copy(const SchedNode &sn) { base_Node::copy(sn);b = sn.getSol(); }
      /// return true if the node is a solution.
      bool isSol() const { return b!=0; }
      /// add a number that represents the number of solution stored by the node.
      void addSol(int _s) { b +=_s; }
      /// returns a reference on the number of solution represented by the node.
      int &getSol() { return b; }
      /// returns a reference on the number of solution represented by the node.
      int getSol() const { return b; }
      /// sets the node as a solution.
      void setSol() { b = 1; }
      /// returns true if the node is a subproblem.
      bool isPb() const { return b==0; }
      /// sets the node as a subproblem (this is the default). 
      void setPb() { b = 0; }
      /// Print the node information
      virtual ostream &Prt(ostream &os) const {
          base_Node::Prt(os);
          os << " solution=" << b; //(b?"true":"false");
          return os;
      }
      /** Fill the strbuff with node information
      * @param st the strbuff to fill.
      */
      virtual void log_space(strbuff<> &st) {
         base_Node::log_space(st);
         st << ","<<(b?"s":"n");
      } 
      /** Fill the strbuff with node information
      * @param st the strbuff to fill.
      */
      virtual void lab_log_space(strbuff<> &st) {
         base_Node::log_space(st);
         st << ","<<"solution or not";
      } 

      /** Pack method to Serialize the SchedNode
       * @param bs the DeSerialize object
       */
      virtual void Pack(Serialize &bs)  const {
         DBGAFF_ENV("SchedNode::Pack","Packing Node");
         base_Node::Pack(bs);bs.Pack(&b,1);
      }
      /** Unpack method to deserialize the SchedNode
       * @param bs the DeSerialize object
       */
      virtual void UnPack(DeSerialize &bs)  {
         DBGAFF_ENV("SchedNode::UnPack","Unpacking Node");
         base_Node::UnPack(bs);bs.UnPack(&b,1);
      }
};


/** The SchedInstance only adds a method to obtain the root node of the search space.
 * The SchedInstance class for a scheduled search inherits from the basic Instance.
 * the functionality that has been added is the SchedInstance::GetRoot() method
 * to obtain the root node of the space search.
 *
  */
template <class Trait> 
class SchedInstance :public base_Instance {
   public :
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
      /// Type of a Node
      typedef typename Trait::Node TheNode;
      /// Type of the Instance
      typedef typename Trait::Instance TheInstance;
      /// Type of a Stat
      typedef typename Trait::Stat TheStat;
      /** Constructor 
        * @param _n a std::string for the identifier
        */
      SchedInstance(std::string &_n) : base_Instance(_n) { 
         DBGAFF_MEM("SchedInstance::SchedInstance","Constructor (para string)");
      }
      /** Constructor 
        * @param _n a char * for the identifier
        */
      SchedInstance(const char *_n) : base_Instance(_n) { 
         DBGAFF_MEM("SchedInstance::SchedInstance","Constructor (para const char *)");
      }
      /** Constructor 
        * @param ia a SchedInstance to copy
        */
      SchedInstance(SchedInstance &ia) : base_Instance(ia) { 
         DBGAFF_MEM("SchedInstance::SchedInstance","Constructor (para Instance &)");
      }
      /** Constructor 
        * @param ia a SchedInstance to copy
        */
      SchedInstance(const SchedInstance &ia) : base_Instance(ia) { 
         DBGAFF_MEM("SchedInstance::SchedInstance","Constructor (para const Instance &)");
      }
      /// Destructor 
      virtual ~SchedInstance()  { 
         DBGAFF_MEM("SchedInstance::~SchedInstance","Destructor");
      }
      /** copy method used by the operator=
        */
      virtual void copy(const SchedInstance &si) { base_Instance::copy(si); }
      /** Operator= for the instance.
        * this operator call the virtual copy method
        */
      const TheInstance &operator=(const TheInstance &ti) { copy(ti); }

      /** Initialize the Instance.
        */
      virtual void Init()=0;
      /** Initialize the search for the algorithm _al.
        * @param _al the algorithm to initialize
        */
      virtual void InitAlgo(TheAlgo *_al) const =0;
      /** Pack method to serialize the SchedInstance
        * @param bs the Serialize object
        */
      virtual void Pack(Serialize &bs) const  {
         DBGAFF_ENV("SchedInstance::Pack","Packing Instance");
         base_Instance::Pack(bs);
      }
      /** Unpack method to deserialize the SchedInstance
        * @param bs the DeSerialize object
        */
      virtual void UnPack(DeSerialize &bs)  {
         DBGAFF_ENV("SchedInstance::UnPack","Unpacking Instance");
         base_Instance::UnPack(bs);
      }
};


/** The GenChild class Generation of the child nodes
 */
template <class Trait>
class SchedGenChild : public base_GenChild {
   public:
      /// Type of a Node
      typedef typename Trait::Node TheNode;
      /// Type of the Info
      typedef typename Trait::Instance TheInstance;
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
  protected:
      /// the algorithm used by the GenChild.
      TheAlgo *algo;
      /// the associated Instance.
      const TheInstance *inst;
   public:
      /** Constructor
      * @param _inst The instance to solve
      * @param _al The associated algorithm 
      */
      SchedGenChild(const TheInstance *_inst, TheAlgo*_al):base_GenChild(),
            algo(_al),inst(_inst) { 
         DBGAFF_MEM("SchedGenChild::SchedGenChild","Constructor (Instance *,Algo *)");
      }
      /// Destructor
      virtual ~SchedGenChild() {
         DBGAFF_MEM("SchedGenChild::~SchedGenChild","Destructor");
      }
      /** operator() which generate the new child nodes and call 
      * the search(nn) method on each of them.
      * @param n the node to explore
      */
      virtual void operator()(TheNode *n) =0;
};


/** The Goal Stat
 */
class GoalStat : public stat {
	public:
		/// Constructor
		GoalStat(bool g=false):stat('G') { 
         add_counter('s',"Generated Solution",g);
         add_counter('u',"Updated Solution",g);
         add_counter('m',"Merge Goal",g);
      }
		/// Destructor
		virtual ~GoalStat() {}
		/// counts the number of calls to the updSol method
		virtual void sol(long l=0) { 
			get_counter('s')++; 
         //get_counter("l")+=l;
		}
		/// counts the number of update of the solution
		virtual void upd(long l=0) { 
			get_counter('u')++; 
         //get_counter("l")-=l;
		}
		/// counts the number of calls to the merge method
		virtual void merge(long l=0) { 
			get_counter('m')++; 
		};

};


/** The abstract Goal for a SchedAlgo
 *  The class is reimplemented for the concrete algorithm in order to control
 *  the search. 
 *  This class could have different implementation according to the aim of 
 *  the search. For example for a Branch and Bound, one want to have the 
 *  optimal solution, to count the number of optimal solution, to have 
 *  the n best soltuions, etc...
 *  The goal of the search must be defined in the trait class for a 
 *  specific problem code.
 *  The main methods are 
 *    - is4Search(Node *n) that tests if a node is suitable to be explored.
 *    - updSol(Node *n) that tests if a node (that must be a solution) is
 *      a candidate to update the best solution known.
 *    - final_merge(Goal *g) performs the final merge.
 *      This method is used at the end of the resoltuion.
 *      This method is only used in parallel environment.
 *    - current_merge(Goal *g) that performs a merge from a Goal given in
 *      parameter, this method is used during the resoltuion. This method is
 *      only used in parallel environment.
 *    - best() returns the best known solution
 *    - worst() returns the worst solution retained by the Goal. Its evaluation
 *      is different from the one from best() only in the case where the Goal
 *      stores solutions with different evaluations.
 */
template<class Trait>
class SchedGoal {
   public:
      /// Type of a Node
      typedef typename Trait::Node TheNode;
      /// Type of the Instance
      typedef typename Trait::Instance TheInstance;
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
      /// Type of the Algo
      typedef typename Trait::Stat TheStat;
      /// Type of the priority
      typedef typename Trait::PriComp ThePriComp;
      /// Type of the GenChild
      typedef typename Trait::GenChild TheGenChild;
      /// Type for the goal of the search
      typedef typename Trait::Goal TheGoal;
      /// Type of the Priority Queue
      typedef PQInterface<TheNode,ThePriComp> ThePQ;
      /** Contructor
        */
      SchedGoal(bool g=false): st(g) {
      }
      /// Destructor
      virtual ~SchedGoal() {
      }
      /** Init method
        * @param ti an Instance object
        */
      virtual void Init(TheInstance *ti) =0;
      /** getStat method 
        */
      GoalStat *getStat() { return &st; }
      /** handleSol method is called for a solution node
        */
      virtual bool updSol(TheNode *n) {
         std::cout <<"Warning : Call SchedGoal updSol() always return false\n";
         return false;
      }
      /** is4Search method
        * @param _n the node to test
        */
      virtual bool is4Search(TheNode *_n) { 
         std::cout <<"Warning : Call SchedGoal i4search() always return true\n";
         return true; 
      }
      /** the current merging method is used to update the solution 
        * with another one, during the execution of algorithms.
        * this method is used in distributed environment.
        * @param _tg the Goal to merge
        */
      virtual bool current_merge(TheGoal *_tg) {
         std::cout <<"Warning : Call SchedGoal current_merge() do nothing\n";
         return true;
      }
      /** the final merging method is used to update the solution 
        * with another one, but at the end of the execution of the algorithm.
        * this method is used in distributed environment.
        * @param _tg the Goal to merge
        */
      virtual bool final_merge(TheGoal *_tg) {
         std::cout <<"Warning : Call SchedGoal final_merge() do nothing\n";
         return true;
      }
      /** operator= for the Goal.
        */
      TheGoal &operator=(TheGoal &_tg) { final_merge(_tg);return this; }

      /** Return the best known solution */
      virtual const TheNode * best() = 0;
      /** Return the worst known solution */
      virtual const TheNode * worst() = 0;
      /** Pack method to Serialize the SchedNode
       * @param bs the DeSerialize object
       */
      virtual void Pack(Serialize &bs) const {
         DBGAFF_ENV("SchedGoal::Pack","Packing Node");
      }
      /** Unpack method to deserialize the SchedNode
       * @param bs the DeSerialize object
       */
      virtual void UnPack(DeSerialize &bs)  {
         DBGAFF_ENV("SchedGoal::UnPack","Unpacking Node");
      }

   protected:
      /** the Statistics of the Goal
        */
      GoalStat st;
};

/** 
 * The SchedStatAlgo represents the statistics for the SchedAlgo
 * it inherits from both StatAlgo and StatPQ.
 */
class SchedStat :public base_Stat {
   public:
      /** Constructor
       * @param n the identifier of the this algorithm
       * @param _id the parent identifier 
       * @param _pqs the Priority queue statistics
       * @param _gst the Goal statistics
       */
      SchedStat(char n,Id *_id,PQStat *_pqs,GoalStat *_gst) : base_Stat(n,_id) {
         add_counter('d',"Deleted Nodes");
         add_counter('i',"Inserted Nodes");
         add_counter('n',"Nonfeasible Nodes");
         add_counter('p',"Pruned Nodes");
         add_time('C',"Genchild Calls");
         add_stat(_pqs);
         add_stat(_gst);
      }
      /// destructor
      virtual ~SchedStat() {
      }
      /// Counts the number of generated nodes
      void StartGC() { get_time('C').start();}
      /// Counts the number of generated nodes
      void EndGC() { get_time('C').end();}
      /// Counts the number of non feasible nodes
      void nofeas() { get_counter('n')++; }
      /// Counts the number of prunned nodes
      void prun() { get_counter('p')++; }
      /// Counts the number of inserted nodes
      void ins() { get_counter('i')++; }
      /// Counts the number of deleted nodes
      void del() { get_counter('d')++; }
      /// Sets the Priority Queue Stats
      void SetPQStat(PQStat *_pqs) { add_stat(_pqs);}
      /// Sets the Goal Stats
      void SetGoalStat(GoalStat *_gst) { add_stat(_gst);}
};


/** The SchedAlgo class
  *
  * This class is used when the search in the space is scheduled.
  *
  * The scheduling could be defined according to a priority (@ref PriGRP), 
  * the class has parametrized type called PriAllocator that has a methods GetPri(Node *) 
  * to obtain a Pri object associated with the Node gien by parameter.
  *
  * The SchedAlgo class define several functionalities
  *   - has the Traitclass as parametrized type
  *   - a GPQ is used to store the pending nodes.
  *   - A PriAllocator (as parametrized type) to create a priority object associated 
  *     to each pending node
  *   - create and use a GenChild object call <b>gc</b> (the exact type is given by Trait class defined as parametrized type) 
  *   - define the Generate() method (which is abstract in this class)
  *   - define the Search() method (which is abstract in this class)
  *   - the funtion class (operator()) which is a template algorithm,  which is mainly a loop 
  *     \verbatim
  *         while the GPQ is not empty
  *            Get the best node n from the GPQ.
  *            Call the Generate method on n to know if n is suitable to be generated
  *            if ok call the gc(n) (the function class associated to the GenChild object gc)  
  *         end while.
  *     \endverbatim
  *     Notice that the operator()(node *n) of the class GenChild must call the Search(Node *c)
  *     method of the algorithm to know what to do with new generated child node c
  *     
  */
template <class Trait>
class SchedAlgo : public base_Algo {
   public:
      /// Type of a Node
      typedef typename Trait::Node TheNode;
      /// Type of the Instance
      typedef typename Trait::Instance TheInstance;
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
      /// Type of the Algo
      typedef typename Trait::Stat TheStat;
      /// Type of the priority
      typedef typename Trait::PriComp ThePriComp;
      /// Type of the GenChild
      typedef typename Trait::GenChild TheGenChild;
      /// Type for the goal of the search
      typedef typename Trait::Goal TheGoal;
      /// Type of the Priority Queue
      typedef PQInterface<TheNode,ThePriComp> ThePQ;

   protected:
      /// the associated Instance.
      const TheInstance *ia;
      /// Pointer to the GPQ object that stores pending nodes.
      ThePQ *gpq;
      /// Pointer to the Goal object that stores the goal to find.
      TheGoal *goal;
      /// the GenChild class
      TheGenChild *gc;
      /// the current parent node
      TheNode *parent;
   public:
      /** Constructor
       * The Genchild used by this algorithm is internaly allocated.
       * @param bsia the instance to solve
       * @param _pq the Priority queue
       * @param tg the Goal of the Search
       * @param st the statistics object
       */
      SchedAlgo(const TheInstance *bsia,ThePQ *_pq,TheGoal *tg,TheStat *st) : 
               base_Algo(st),ia(bsia),gpq(_pq),goal(tg),
               gc(new TheGenChild(ia,(TheAlgo *)this)),parent(0) { 
         BOBASSERT(gpq!=0,NullPointerException,"SchedAlgo","GPQ is null");
         DBGAFF_MEM("SchedAlgo::SchedAlgo","Constructor (Instance *,PQ *,Stat *)");
      }
      /** Destructor 
        * The Genchild is deleted
        */
      virtual ~SchedAlgo() {
         delete gc;gc=0;
         DBGAFF_MEM("SchedAlgo::~SchedAlgo","Destructor");
      }
      /// the run method that performs the main loop.
      virtual void operator()() {
         TheNode *n;
         BOBASSERT(base_Algo::isEtInit() || base_Algo::isEtResumed(),StateException,
                     "SchedAlgo::Operator()"," This object has not been initialized ");
         base_Algo::StStart();
         base_Algo::toRun();
         GPQ()->addRef();
         DBGAFF_ALGO("SchedAlgo::operator()","Before the main loop");
         while ( base_Algo::isEtRun() && (n=GPQGet())!=0 ) {
            DBGAFF_ALGO("SchedAlgo::operator()",
                  "We have a node and the state of the algo is ok");
            if ( !goal->is4Search(n) ) {
               DBGAFF_ALGO("SchedAlgo::operator()","Node must be discarded");
               if ( n->isDel() ) delete n;
               getStat()->prun();
               continue;
            }
            parent=n;
            getStat()->StartGC();
            DBGAFF_ALGO("SchedAlgo::operator()","Call the genchild");
            (*gc)(n);
            getStat()->EndGC();
            if ( n->isDel() ) delete n;
         }
         GPQ()->remRef();
         base_Algo::StEnd();
         DBGAFF_ALGO("SchedAlgo::operator()","ends !");
      }
      /// Ends the algorithm 
      virtual void End() { 
            DBGAFF_ALGO("SchedAlgo::end()","is called");
      }
      /** The Search method 
        * This method is called by the GenChild::operator(), for each new 
        * generated nodes that are suitable to be explorated.
        * The actual code is simply an insertion of the Node in the GPQ.
        * This methods could be redefined for a specific algorithm.
        * @param n the node to search
        */
      virtual void Search(TheNode *n) {
         log_space(n);
         if ( n->isSol() ) {
            goal->updSol(n);
            if ( n->isDel() ) delete n;
            return;
         }
         if ( ! goal->is4Search(n) ) {
            DBGAFF_ALGO("SchedAlgo::Search()","Goal wants to discard node");
            if ( n->isDel() ) delete n;
            getStat()->prun();
            return;
         }
         DBGAFF_ALGO("SchedAlgo::Search()","Insert the node");
         GPQIns(n);
      };
      /**
         @name GPQ Access methods.
      */
      //@{
      /// get the Global priority Queue
      virtual ThePQ *&GPQ() { return gpq; }
      /// inserts a EvalNode in the GPQ
      void GPQIns(TheNode *n) {
         getStat()->ins();
         GPQ()->Ins(n);
      }
      /// gets  a EvalNode in the GPQ
      TheNode *GPQGet() {
        getStat()->del();
        return (TheNode *)GPQ()->Del();
      }
      //@}

      /// gets the Instance
      const TheInstance *getInstance() { return ia; }
      /// gets the Instance
      TheGoal *getGoal() { return goal; }
      /// gets the GenChild
      TheGenChild *getGC() { return gc; }

      /**
         @name Statistics methods.
      */
      //@{
      /// return the statistics with the right types
      TheStat *getStat() const { return (TheStat *)base_Algo::getBStat(); }
      /// statisitics for non feasible nodes
      void StNoFea() { getStat()->nofeas(); }
      /// statisitics for non feasible nodes
      void StPrun() { getStat()->prun(); }
      /// log the node an its parent.
      void log_space(TheNode *n) {
         #ifdef ENABLE_SPACE
         strbuff<> st;
         n->id()=core::get_new_id();
         st << getInstance()->id()->str()<< ":N ("<<n->id()<<",";
         if ( parent!=0 ) st << parent->id();
         else st << "-";
         n->log_space(st);
         st <<")\n";
         core::log().out(true,st);
         #endif
      }
      //@}
};


//@}
};
#endif

